import tarfile
import sys
import os
import subprocess
import time
import traceback
import json
import pip

def GetTmFmt():
    Tm = time.localtime()
    return "%u-%02u-%02u %02u:%02u:%02u" % (Tm.tm_year, Tm.tm_mon, Tm.tm_mday, Tm.tm_hour, Tm.tm_min, Tm.tm_sec)

# program must be run as sudo

def dctMerge(dct0, dct1, chooseFn):
    rtn = {}
    both = set(dct0)&set(dct1)
    only_dct0 = set(dct0)-set(dct1)
    only_dct1 = set(dct1)-set(dct0)
    for k in both: rtn[k] = chooseFn(k, dct0[k], dct1[k])
    for k in only_dct0: rtn[k] = dct0[k]
    for k in only_dct1: rtn[k] = dct1[k]
    return rtn

def metainf_merger(k, val0, val1):
    if k == "version": return val0
    elif k == "os_pkg_ver": return val1
    else: return val0

def convFName(fName):
    fName = "/".join(filter(len, fName.replace("\\", "/").split("/")))
    if os.name == "posix": fName = "/" + fName
    return fName
def convDName(dName):
    dName = "/".join(filter(len, dName.replace("\\", "/").split("/"))) + "/"
    if os.name == "posix": dName = "/" + dName
    return dName

baseDir = convFName(__file__).rsplit("/", 1)[0] + "/"

#update archive file structure
#/
#/metainf.json
#/apt-offline.zip
#/pip/
#/pip/requirements.txt
#/pip/*.tar.gz <any number of occurences>
#/pip/*.zip <any number of occurences>
def main(argv):
    logFl = open(baseDir + "OS_Updater.log", "wb")
    def Log(Str):
        logFl.write(GetTmFmt() + " [OS_Updater] " + Str + "\n")
    try:
        assert len(argv) >= 3, "Usage %s <update file> <extraction point>" % (
            "OS_Updater.py" if len(argv) == 0 else argv[0].replace("\\", "/").rsplit("/")[-1])
        assert os.name == "posix", "must be run on posix operating system not '%s'" % os.name
        End = 3
        dctWordOpts = {
            "umount": None,
            "wait": None
        }
        chToWord = {
            "u": "umount",
            "w": "wait"
        }
        while len(argv) > End:
            if argv[End].startswith("--"):
                word = argv[End][2:]
                if word not in dctWordOpts:
                    Log("Unrecognized option: --%s" % word)
                    return
                End += 1
                if len(argv) <= End:
                    Log("expected argument for option --%s" % word)
                    return
                dctWordOpts[word] = argv[End]
                End += 1
            elif argv[End].startswith("-"):
                ch = argv[End][1:]
                if len(ch) != 1:
                    Log("expected single character option after -, -%s" % ch)
                    return
                End += 1
                if len(argv) <= End:
                    Log("expected argument for option -%s" % ch)
                    return
                word = chToWord[ch]
                dctWordOpts[word] = argv[End]
                End += 1
            else:
                break

        update_file = convFName(argv[1])
        extraction_point = convDName(argv[2])
        assert os.path.isfile(update_file), "update file must be a file"
        if os.path.exists(extraction_point):
            assert os.path.isdir(extraction_point), "extraction point must be directory"
        else:
            os.mkdir(extraction_point)
        if dctWordOpts["wait"] is not None:
            pid = int(dctWordOpts["wait"])
            waitTxt = "waiting on %u"%pid
            if os.name == "posix":
                waitDir = "/proc/%u/stat" % pid
                while os.path.exists(waitDir):
                    Log(waitTxt)
                    time.sleep(1)
                Log("Done waiting")
            elif os.name == "nt":
                import ctypes
                kernel32 = ctypes.windll.kernel32
                kernel32.OpenProcess.argtypes = [ctypes.c_ulong, ctypes.c_int, ctypes.c_ulong]
                kernel32.OpenProcess.restype = ctypes.c_void_p
                Log(waitTxt)
                try:
                    hProc = kernel32.OpenProcess(0x00100400, 0, pid)
                    if hProc is not None and hProc != 0:
                        os.waitpid(hProc, 0)
                        kernel32.CloseHandle(hProc)
                except OSError:
                    pass
                time.sleep(.25)
                Log("Done waiting")
        metainf = {}
        try:
            with open(baseDir + "metainf.json", "r") as fl:
                metainf = json.load(fl)
        except (OSError, IOError):
            metainf = {}
        except ValueError:
            metainf = {}
            # mal-formed metainf.json
        f_metainf = {}
        fl0 = tarfile.open(argv[1])
        metainf_member = None
        try:
            metainf_member = fl0.getmember("metainf.json")
        except KeyError:
            # missing metainf.json
            raise
        fl = fl0.extractfile(metainf_member)
        try:
            f_metainf = json.load(fl)
        except ValueError:
            # mal-formed metainf.json
            raise
        finally:
            fl.close()
        apt_fName = "apt-offline.zip"
        fl0.extract(fl0.getmember(apt_fName), extraction_point)
        res = subprocess.call(["apt-offline", "install", extraction_point + apt_fName])
        if res != 0:
            raise ValueError(res, "Failed to run apt-offline")
        pip_members = []
        for member in fl0.getmembers():
            if member.name.split("/")[0] == "pip":
                pip_members.append(member)
        fl0.extractall(extraction_point, pip_members)
        pip_point = extraction_point + "pip/"
        pip.main(["install", "--no-index", "--find-links", "file://" + pip_point, "-r", pip_point + "requirements.txt"])
        fl0.close()
        metainf = dctMerge(metainf, f_metainf, metainf_merger)
        with open(baseDir + "metainf.json", "w") as fl:
            json.dump(metainf, fl)
        if dctWordOpts["umount"] is not None:
            arg = dctWordOpts["umount"]
            if os.name == "posix":
                subprocess.call(["sudo", "umount", arg])
            else:
                Log("Other operating system types (not posix) may not be able to dismount, doing nothing")
        if len(argv) > End:
            subprocess.Popen(argv[End:], close_fds=True)
    except:
        Log("ERROR: " + traceback.format_exc()[:-1])
    else:
        Log("SUCCESS")
    finally:
        Log("Exitting OS_Updater")
        logFl.close()


if __name__ == "__main__": main(sys.argv)
