from os.path import isfile
from os import remove, rename
from sys import stderr, stdout
from time import localtime
from threading import Lock


# get time string for use in log files or text data
def GetTmFmt():
    Tm = localtime()
    return "%u-%02u-%02u %02u:%02u:%02u" % (Tm.tm_year, Tm.tm_mon, Tm.tm_mday, Tm.tm_hour, Tm.tm_min, Tm.tm_sec)


# get time string for use in filenames
def fNameSafeGetTmFmt():
    Tm = localtime()
    return "%u-%02u-%02u_%02u.%02u.%02u" % (Tm.tm_year, Tm.tm_mon, Tm.tm_mday, Tm.tm_hour, Tm.tm_min, Tm.tm_sec)


class Logger(object):
    LOG_LVLS = ["SEVERE", "ERROR", "WARNING", "INFO", "FINE", "FINER", "FINEST"]
    __slots__ = ["modName", "lstOuts", "logLk", "minLvl", "maxLvl"]

    def __init__(self, modName, lstOuts, logLk=None):
        """
        :param str|unicode modName:
        :param list[(file, int, int)] lstOuts:
        :param Lock|None logLk:
        """
        self.modName = modName
        self.lstOuts = lstOuts
        self.logLk = logLk if logLk is not None else Lock()
        self.minLvl = min(lstOuts, key=lambda x: x[1])[1]
        self.maxLvl = max(lstOuts, key=lambda x: x[2])[2]

    def Log(self, s, lvl=3):
        """
        :param str|unicode s:
        :param int lvl:
        """
        if self.minLvl > lvl or self.maxLvl < lvl:
            return
        tm = localtime()
        sOut = (
            u"%u-%02u-%02u %02u:%02u:%02u [%s][%s] %s\n" % (
                tm.tm_year, tm.tm_mon, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec,
                self.LOG_LVLS[lvl], self.modName, s
            )
        ).encode("utf-8")
        with self.logLk:
            for fl, a, b in self.lstOuts:
                if a <= lvl <= b:
                    fl.write(sOut)
                    fl.flush()

    def SpawnLog(self, modName, lstOutLvls=None):
        """
        :param str|unicode modName:
        :param list[(int, int)|None]|None lstOutLvls:
        """
        lstOuts = self.lstOuts
        logLk = self.logLk
        if lstOutLvls is None:
            return Logger(modName, lstOuts, logLk)
        assert len(lstOutLvls) <= len(lstOuts)
        lstOuts1 = list(lstOuts)
        for c, x in enumerate(lstOutLvls):
            if x is not None:
                lstOuts1[c] = (lstOuts1[c][0], x[0], x[1])
        return Logger(modName, lstOuts1, logLk)


class LoggingContext(object):
    __slots__ = ["baseDir", "name", "fNameLog", "crashBegin", "crashEnd"]

    def __init__(self, baseDir, name, crashBegin="crash-", crashEnd=".log"):
        """
        :param str|unicode baseDir:
        :param str|unicode name:
        :param str|unicode crashBegin:
        :param str|unicode crashEnd:
        """
        self.baseDir = baseDir
        self.name = name
        assert len(baseDir) == 0 or baseDir.endswith("/")
        self.fNameLog = baseDir + name
        self.crashBegin = crashBegin
        self.crashEnd = crashEnd

    def createAndSwapLogFile(self, modName, minLvlFl=None, maxLvlFl=None, lstOuts=None):
        """
        :param str|unicode modName:
        :param int|None minLvlFl:
        :param int|None maxLvlFl:
        :param list[(file, int, int)]|None lstOuts:
        """
        fNameLog = self.fNameLog
        if isfile(fNameLog):
            oldName = fNameLog + ".old"
            if isfile(oldName):
                remove(oldName)
            while True:
                try:
                    rename(fNameLog, oldName)
                except WindowsError as Exc:
                    if Exc.winerror != 32: raise
                else:
                    break
        if minLvlFl is None:
            minLvlFl = 0
        if maxLvlFl is None:
            maxLvlFl = len(Logger.LOG_LVLS) - 1
        if lstOuts is None:
            lstOuts = [(stderr, 0, 2), (stdout, 3, 3)]
        fl = open(fNameLog, "wb")
        return fl, Logger(modName, lstOuts + [(fl, minLvlFl, maxLvlFl)])

    def closeLog(self, fl, logger, lstLastLog=None, isCrash=False):
        """
        :param file fl:
        :param Logger logger:
        :param list[str|unicode|(str|unicode, int)]|None lstLastLog:
        :param bool isCrash:
        """
        crashName = None
        if isCrash:
            tm = localtime()
            crashName = "%s%s%u-%02u-%02u_%02u.%02u.%02u%s" % (
                self.baseDir, self.crashBegin,
                tm.tm_year, tm.tm_mon, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec,
                self.crashEnd
            )
        if lstLastLog is not None:
            for s in lstLastLog:
                if isinstance(s, (str, unicode)):
                    s = (s, 3)
                logger.Log(*s)
        fl.close()
        if crashName is not None:
            rename(self.fNameLog, crashName)